#include "libnazgul.h"
#include "nzg_ids.h"

/* TODO: remplacer le bool de msgPoolData par un identifiant
   de semaphore. Le semaphore contient poolNb valeurs (et indique
   donc le nombre de ressources disponnibles).

*/

#define ANYPOOL -1
#define SPECIFICPOOL 0

void * msgAllocate(msgSpace *space,
	int pool,
	int taille,
	int option
	){
  void * resultAddr=NULL;
  int i, mSPoolDataTabFd;
  msgPoolId resultPoolId;
  /* tableau des valeurs des semPoolCoef/pool pour identifier le pool
   * qui sera libr le plus rapidement */
  float semPoolCoef[space->poolNb]; 
  int idxPoolOptimum;
  bool gotRessourceSem;
  sem_t * poolDataTabSemFd=NULL;
  msgPoolDataTabSemId ressourceSemId;
  sem_t * ressourceSemFd=NULL;
  int * ressourceSemVal=NULL;
  float minPoolCoef;
  int selectedIdx;
  
  selectedIdx=-1;

  /* TODO: verifier le premier arg du shm_open */
  mSPoolDataTabFd=shm_open(space->poolDataTabId,O_RDWR,MSGSPACE_DEFAULT_MODE);
  if (mSPoolDataTabFd == -1 ) {
	fprintf( stderr, "Allocate %s failed: %s\n",
		(char*)space->poolDataTabId,
		strerror( errno ) );
	return NULL;
  }

  msgPoolData * mSPoolDataTab;

  mSPoolDataTab = mmap( 0, (space->poolNb) * sizeof( msgPoolData ),
	  PROT_READ | PROT_WRITE,
	  MAP_SHARED, mSPoolDataTabFd, 0 );

  if( mSPoolDataTab == MAP_FAILED) {
	fprintf( stderr, "mmap failed: %s\n",
		strerror( errno ) );
	return NULL;
  }

	gotRessourceSem=false;

  /* initialisation des coefs */
  for (i=0;i<(space->poolNb);i++){
	semPoolCoef[i]=-1;
  }
  
  int nbLockedSem=0;
  if ( pool == ANYPOOL){
	// choisir le pool au hasard (ou presque)
	for(i=0; i<(space->poolNb); i++) {
	  if(mSPoolDataTab[i].bufferSize >= taille) {
		/* choisir le numero du semaphore
		   en fonction du nombre de lock poses / nombre de buffer */
		msgPoolSemIdIntern(ressourceSemId,space->id,i);
		ressourceSemFd = sem_open(ressourceSemId,O_CREAT,SEM_DEFAULT_MODE,0);
		if (ressourceSemFd == SEM_FAILED){ 
#warning "Grer mieu les erreurs..." 
		  perror("sem_open"); return NULL; 
		}
		/* on remplit le tableau avec les valeurs des semaphores */
				
		if (sem_getvalue(ressourceSemFd, ressourceSemVal) < 0){ 
#warning "Grer mieu les erreurs..."
		  perror("sem_getvalue"); return NULL; 
		}

		if ((*ressourceSemVal) < 0){
		  /* il y a ressourceSemVal processus qui attendent dja... */
		  semPoolCoef[nbLockedSem] = 
			(float) (- (*ressourceSemVal) / mSPoolDataTab[i].bufferNb);
		  nbLockedSem++;
		}
		if(sem_trywait(ressourceSemFd)) {
		  /* choisir la 1ere pool de taille plus grande 
		   * libre si possible */
		  gotRessourceSem=true;
		  selectedIdx=i;
		  break;
		}
	if(	sem_close(ressourceSemFd) <0){
#warning "Grer mieu les erreurs..." 
		  perror("sem_getvalue"); return NULL; 
	}
	  } // if buffSize > taille
	} // for

	if (!gotRessourceSem) {		
	  minPoolCoef= semPoolCoef[0];
	  idxPoolOptimum = 0;
	  /* on cherche le pool avec le moins de lock poses / nbre de buffer
	   * le numro du pool est stock dans idxPoolOptimum */
	  for(i=0; i<nbLockedSem; i++) {
		if ((semPoolCoef[i] != -1) 
			&& (semPoolCoef[i] < minPoolCoef)) {
		  minPoolCoef = semPoolCoef[i];
		  idxPoolOptimum = i;
		}
	  }

	  if (minPoolCoef == -1){
		/* il n'y a aucune pool dont la taille satisfait la demande */
		return NULL;
	  } else {
		selectedIdx=idxPoolOptimum;
	  }
	}
  }else {
	selectedIdx=pool;
  }

  if (!gotRessourceSem){
	strcpy(resultPoolId,mSPoolDataTab[selectedIdx].id);
	msgPoolSemIdIntern(ressourceSemId,space->id,selectedIdx);
	ressourceSemFd=sem_open(ressourceSemId,O_CREAT,SEM_DEFAULT_MODE,0);
	if(ressourceSemFd==SEM_FAILED){
	  perror("sem_open");
	  return NULL;
	}
  }

  if (sem_wait(ressourceSemFd) < 0){
	perror("sem_wait");
	sem_close(ressourceSemFd);
	return NULL;
  }		

  /* on a acqui un semaphore pour la ressouce */
  /* on acquiert le droit de modifier les infos sur la ressource */
  poolDataTabSemFd=sem_open(space->poolDataTabSemId,O_CREAT,SEM_DEFAULT_MODE,1);

  /* on modifie maintenant les donnes */

  /* TODO: mapper le buffer libre dans l'esp addr du proc */

  sem_close(poolDataTabSemFd);

  /* TODO: unmapper le msgPoolDataTab */


  return resultAddr;
}
